@@ -76,25 +76,18 @@ static const char * make_expression(Dictionary dict,
 				"Missing direction character in connector string: %s", con_start);
 
 		/* Create an expression to hold the connector */
-		Exp* e = Exp_create(dict->Exp_pool);
-		e->dir = *p;
-		e->type = CONNECTOR_type;
-		e->operand_next = NULL;
-		e->cost = 0.0;
 		char * constr = NULL;
+		bool multi = false;
 		if ('@' == *con_start)
 		{
 			constr = strndupa(con_start+1, p-con_start-1);
-			e->multi = true;
+			multi = true;
 		}
 		else
-		{
 			constr = strndupa(con_start, p-con_start);
-			e->multi = false;
-		}
 
-		e->condesc = condesc_add(&dict->contable,
-		                           string_set_add(constr, dict->string_set));
+		Exp* e = make_connector_node(dict, dict->Exp_pool, constr, *p, multi);
+
 		*pex = e;
 	}
 
@@ -125,14 +118,7 @@ static const char * make_expression(Dictionary dict,
 	assert(NULL != rest, "Badly formed expression %s", exp_str);
 
 	/* Join it all together. */
-	Exp* join = Exp_create(dict->Exp_pool);
-	join->type = etype;
-	join->operand_next = NULL;
-	join->cost = 0.0;
-
-	join->operand_first = *pex;
-	(*pex)->operand_next = rest;
-	rest->operand_next = NULL;
+	Exp* join = make_join_node(dict->Exp_pool, etype, *pex, rest);
 
 	*pex = join;
 
@@ -177,7 +163,7 @@ static int exp_cb(void *user_data, int argc, char **argv, char **colName)
 	make_expression(dict, argv[0], &exp);
 	assert(NULL != exp, "Failed expression %s", argv[0]);
 
-	if (!strtodC(argv[1], &exp->cost))
+	if (!strtofC(argv[1], &exp->cost))
 	{
 		prt_error("Warning: Invalid cost \"%s\" in expression \"%s\" "
 		          "(using 1.0)\n", argv[1], argv[0]);
@@ -194,13 +180,7 @@ static int exp_cb(void *user_data, int argc, char **argv, char **colName)
 	/* If the second expression, OR-it with the existing expression. */
 	if (OR_type != bs->exp->type)
 	{
-		Exp* orn = Exp_create(dict->Exp_pool);
-		orn->type = OR_type;
-		orn->cost = 0.0;
-
-		orn->operand_first = exp;
-		exp->operand_next = bs->exp;
-		bs->exp = orn;
+		bs->exp = make_or_node(dict->Exp_pool, exp, bs->exp);
 		return 0;
 	}
 
@@ -506,7 +486,7 @@ static void add_categories(Dictionary dict)
 
 		dict->category[i].num_words = bs.count;
 		dict->category[i].word =
-			malloc(bs.count * sizeof(dict->category[0].word));
+			malloc(bs.count * sizeof(*dict->category[0].word));
 
 		/* ------------------ */
 		/* For each category, get the (subscripted) words in the category */
@@ -606,10 +586,15 @@ Dictionary dictionary_create_from_db(const char *lang)
 	dict->lookup_list = db_lookup_list;
 	dict->lookup_wild = db_lookup_wild;
 	dict->free_lookup = db_free_llist;
-	dict->lookup = db_lookup;
+	dict->exists_lookup = db_lookup;
+	dict->clear_cache = dict_node_noop;
 	dict->close = db_close;
+
+	dict->dynamic_lookup = true;
 	condesc_init(dict, 1<<8);
 
+	dict->dfine.set = string_id_create();
+
 	dict->Exp_pool = pool_new(__func__, "Exp", /*num_elements*/4096,
 	                          sizeof(Exp), /*zero_out*/false,
 	                          /*align*/false, /*exact*/false);
