@@ -26,6 +26,7 @@ struct Tconnector_struct
 {
 	Tconnector * next;
 	Exp *e; /* a CONNECTOR_type element from which to get the connector  */
+	Connector *tracon; /* the created tracon, set through memory sharing */
 };
 
 typedef struct clause_struct Clause;
@@ -82,18 +83,17 @@ static Tconnector * catenate(Tconnector * e1, Tconnector * e2, Pool_desc *tp)
 {
 	Tconnector head;
 	Tconnector *preve = &head;
-	Tconnector *newe = &head;
 
 	for (;e1 != NULL; e1 = e1->next)
 	{
-		newe = pool_alloc(tp);
+		Tconnector *newe = pool_alloc(tp);
 		*newe = *e1;
 
 		preve->next = newe;
 		preve = newe;
 	}
 
-	preve->next = e2;
+	preve->next = e2; /* tracon memory sharing */
 	return head.next;
 }
 
@@ -106,6 +106,7 @@ static Tconnector * build_terminal(Exp *e, clause_context *ct)
 	c->e = e;
 	c->next = NULL;
 	c->e->pos = ct->exp_pos++;
+	c->tracon = NULL;
 	return c;
 }
 
@@ -148,13 +149,14 @@ static Clause * build_clause(Exp *e, clause_context *ct, Clause **c_last)
 				for (c4 = c2; c4 != NULL; c4 = c4->next)
 				{
 					float maxcost = MAX(c3->maxcost,c4->maxcost);
-					if (maxcost + e->cost > ct->cost_cutoff) continue;
+					/* Cannot use this shortcut due to negative costs. */
+					//if (maxcost + e->cost > ct->cost_cutoff) continue;
 
 					c = pool_alloc(ct->Clause_pool);
 					if ((c_head == NULL) && (c_last != NULL)) *c_last = c;
 					c->cost = c3->cost + c4->cost;
 					c->maxcost = maxcost;
-					c->c = catenate(c3->c, c4->c, ct->Tconnector_pool);
+					c->c = catenate(c4->c, c3->c, ct->Tconnector_pool);
 					c->next = c_head;
 					c_head = c;
 				}
@@ -232,6 +234,11 @@ build_disjunct(Sentence sent, Clause * cl, const char * string,
 {
 	Disjunct *dis, *ndis;
 	Pool_desc *connector_pool = NULL;
+	bool sat_solver = false;
+
+#if USE_SAT_SOLVER
+		sat_solver = (opts != NULL) && opts->use_sat_solver;
+#endif /* USE_SAT_SOLVER */
 
 	dis = NULL;
 	for (; cl != NULL; cl = cl->next)
@@ -240,7 +247,7 @@ build_disjunct(Sentence sent, Clause * cl, const char * string,
 		if (cl->maxcost > cost_cutoff) continue;
 
 #if USE_SAT_SOLVER
-		if (opts->use_sat_solver) /* For the SAT-parser, until fixed. */
+		if (sat_solver) /* For the SAT-parser, until fixed. */
 		{
 			ndis = xalloc(sizeof(Disjunct));
 		}
@@ -253,22 +260,31 @@ build_disjunct(Sentence sent, Clause * cl, const char * string,
 		ndis->left = ndis->right = NULL;
 
 		/* Build a list of connectors from the Tconnectors. */
+		Connector **jet[2] = { &ndis->left, &ndis->right };
+		bool is_tracon[2] = { false, false };
 		for (Tconnector *t = cl->c; t != NULL; t = t->next)
 		{
+			int idir = ('+' == t->e->dir);
+
+			if (is_tracon[idir]) continue; /* this direction is complete */
+			if (NULL != t->tracon)
+			{
+				/* Use the cashed tracon and mark this direction as complete. */
+				*(jet[idir]) = t->tracon;
+				is_tracon[idir] = true;
+				continue;
+			}
+
 			Connector *n = connector_new(connector_pool, t->e->condesc, opts);
-			Connector **loc = ('-' == t->e->dir) ? &ndis->left : &ndis->right;
+			t->tracon = n; /* cache this tracon */
 
 			n->exp_pos = t->e->pos;
 			n->multi = t->e->multi;
 			n->farthest_word = t->e->farthest_word;
-			n->next = *loc;   /* prepend the connector to the current list */
-			*loc = n;         /* update the connector list */
-		}
 
-		bool sat_solver = false;
-#if USE_SAT_SOLVER
-		sat_solver = opts->use_sat_solver;
-#endif /* USE_SAT_SOLVER */
+			*(jet[idir]) = n;
+			jet[idir] = &n->next;
+		}
 
 		/* XXX add_category() starts category strings by ' '.
 		 * FIXME Replace it by a better indication. */
@@ -308,7 +324,7 @@ Disjunct *build_disjuncts_for_exp(Sentence sent, Exp* exp, const char *word,
 	clause_context ct = { 0 };
 
 	ct.cost_cutoff = cost_cutoff;
-	if (sent->Clause_pool == NULL)
+	if (unlikely(sent->Clause_pool == NULL))
 	{
 		ct.Clause_pool = pool_new(__func__, "Clause",
 		                          /*num_elements*/4096, sizeof(Clause),
